GRAPH_STYLING = dict(
    fontname="Courier",
    fontsize="12",
    style="dotted",
    labeljust="l",
)
IN_ARG_STYLING = dict(
    shape="box",
    color="green",
    fontname="Courier",
    fontsize="10",
)
LITERAL_STYLING = dict(
    shape="box",
    color="orange",
    fontname="Courier",
    fontsize="10",
)
CONST_ARG_STYLING = dict(
    shape="box",
    color="darkgreen",
    fontname="Courier",
    fontsize="10",
)
OUT_ARG_STYLING = dict(
    shape="box",
    color="red",
    fontname="Courier",
    fontsize="10",
)
VAR_STYLING = dict(
    shape="box",
    color="blue",
    fontname="Courier",
    fontsize="10",
)
PRIMITIVE_STYLING = dict(
    shape="box",
    fontname="Courier",
    fontsize="10",
    color="grey",
)
COND_NODE_STYLING = dict(
    shape="box",
    fontname="Courier",
    fontsize="10",
    color="grey",
)
FUNCTION_NODE_STYLING = dict(
    shape="rectangle", fontname="Courier", fontsize="10", style="dotted"
)
